@@ -37,12 +37,14 @@ from sample_factory.algorithms.utils.algo_utils import EXTRA_PER_POLICY_SUMMARIE
 from sample_factory.envs.env_utils import get_default_reward_shaping
 from sample_factory.utils.timing import Timing
 from sample_factory.utils.utils import summaries_dir, experiment_dir, log, str2bool, memory_consumption_mb, cfg_file, \
-    ensure_dir_exists, list_child_processes, kill_processes, AttrDict, done_filename, save_git_diff
+    ensure_dir_exists, list_child_processes, kill_processes, AttrDict, done_filename, save_git_diff, init_file_logger
 
 if os.name == 'nt':
     from sample_factory.utils import Queue as MpQueue
 else:
     from faster_fifo import Queue as MpQueue
+    # noinspection PyUnresolvedReferences
+    import faster_fifo_reduction
 
 
 torch.multiprocessing.set_sharing_strategy('file_system')
@@ -121,6 +123,10 @@ class APPO(ReinforcementLearningAlgorithm):
         p.add_argument('--exploration_loss_coeff', default=0.003, type=float,
                        help='Coefficient for the exploration component of the loss function.')
         p.add_argument('--value_loss_coeff', default=0.5, type=float, help='Coefficient for the critic loss')
+        p.add_argument('--kl_loss_coeff', default=0.0, type=float,
+                       help='Coefficient for fixed KL loss (as used by Schulman et al. in https://arxiv.org/pdf/1707.06347.pdf). '
+                            'Highly recommended for environments with continuous action spaces.',
+                       )
         p.add_argument('--exploration_loss', default='entropy', type=str, choices=['entropy', 'symmetric_kl'],
                        help='Usually the exploration loss is based on maximizing the entropy of the probability'
                             ' distribution. Note that mathematically maximizing entropy of the categorical probability '
@@ -159,10 +165,9 @@ class APPO(ReinforcementLearningAlgorithm):
                  'policy-lag may exceed this value.',
         )
         p.add_argument(
-            '--min_traj_buffers_per_worker', default=2, type=int,
-            help='How many shared rollout tensors to allocate per actor worker to exchange information between actors and learners'
-                 'Default value of 2 is fine for most workloads, except when differences in 1-step simulation time are extreme, like with some DMLab environments.'
-                 'If you see a lot of warnings about actor workers having to wait for trajectory buffers, try increasing this to 4-6, this should eliminate the problem at a cost of more RAM.',
+            '--traj_buffers_excess_ratio', default=1.3, type=float,
+            help='Increase this value to make sure the system always has enough free trajectory buffers (can be useful when i.e. a lot of inactive agents in multi-agent envs)'
+                 'Decrease this to 1.0 to save as much RAM as possible.',
         )
         p.add_argument(
             '--decorrelate_experience_max_seconds', default=10, type=int,
@@ -316,6 +321,7 @@ class APPO(ReinforcementLearningAlgorithm):
     def initialize(self):
         self._save_cfg()
         save_git_diff(experiment_dir(cfg=self.cfg))
+        init_file_logger(experiment_dir(self.cfg))
 
     def finalize(self):
         pass
