@@ -17,11 +17,20 @@ import numpy as np
 import psutil
 from colorlog import ColoredFormatter
 
-ch = logging.StreamHandler()
-ch.setLevel(logging.DEBUG)
 
-formatter = ColoredFormatter(
-    "%(log_color)s[%(asctime)s][%(process)05d] %(message)s",
+# Logging
+
+log = logging.getLogger('rl')
+log.setLevel(logging.DEBUG)
+log.handlers = []  # No duplicated handlers
+log.propagate = False  # workaround for duplicated logs in ipython
+log_level = logging.DEBUG
+
+stream_handler = logging.StreamHandler()
+stream_handler.setLevel(log_level)
+
+stream_formatter = ColoredFormatter(
+    '%(log_color)s[%(asctime)s][%(process)05d] %(message)s',
     datefmt=None,
     reset=True,
     log_colors={
@@ -35,13 +44,16 @@ formatter = ColoredFormatter(
     secondary_log_colors={},
     style='%'
 )
-ch.setFormatter(formatter)
+stream_handler.setFormatter(stream_formatter)
+log.addHandler(stream_handler)
 
-log = logging.getLogger('rl')
-log.setLevel(logging.DEBUG)
-log.handlers = []  # No duplicated handlers
-log.propagate = False  # workaround for duplicated logs in ipython
-log.addHandler(ch)
+
+def init_file_logger(experiment_dir_):
+    file_handler = logging.FileHandler(join(experiment_dir_, 'sf_log.txt'))
+    file_handler.setLevel(log_level)
+    file_formatter = logging.Formatter(fmt='[%(asctime)s][%(process)05d] %(message)s', datefmt=None, style='%')
+    file_handler.setFormatter(file_formatter)
+    log.addHandler(file_handler)
 
 
 # general Python utilities
@@ -117,14 +129,18 @@ def safe_get(q, timeout=1e6, msg='Queue timeout'):
 
 
 def safe_put(q, msg, attempts=3, queue_name=''):
+    safe_put_many(q, [msg], attempts, queue_name)
+
+
+def safe_put_many(q, msgs, attempts=3, queue_name=''):
     for attempt in range(attempts):
         try:
-            q.put(msg)
+            q.put_many(msgs)
             return
         except Full:
-            log.warning('Could not put msg to queue, the queue %s is full! Attempt %d', queue_name, attempt)
+            log.warning('Could not put msgs to queue, the queue %s is full! Attempt %d', queue_name, attempt)
 
-    log.error('Failed to put msg to queue %s after %d attempts. The message is lost!', queue_name, attempts)
+    log.error('Failed to put msgs to queue %s after %d attempts. Messages are lost!', queue_name, attempts)
 
 
 # CLI args
@@ -226,15 +242,21 @@ def list_child_processes():
 
 
 def kill_processes(processes):
+    # do not kill to avoid permanent memleaks
+    # https://pytorch.org/docs/stable/multiprocessing.html#file-system-file-system
+    processes_to_save = ['torch_shm', 'resource_tracker', 'semaphore_tracker']
+
     for p in processes:
         try:
-            if 'torch_shm' in p.name():
-                # do not kill to avoid permanent memleaks
-                # https://pytorch.org/docs/stable/multiprocessing.html#file-system-file-system
+            kill_proc = True
+            for proc_to_save in processes_to_save:
+                if any(proc_to_save in s for s in [p.name()] + p.cmdline()):
+                    kill_proc = False
+
+            if not kill_proc:
                 continue
 
-            # log.debug('Child process name %d %r %r %r', p.pid, p.name(), p.exe(), p.cmdline())
-            log.debug('Child process name %d %r %r', p.pid, p.name(), p.exe())
+            log.debug('Child process name %d %r %r %r', p.pid, p.name(), p.exe(), p.cmdline())
             if p.is_running():
                 log.debug('Killing process %s...', p.name())
                 p.kill()
